#include "rtweekend.h"

#include "hittable_list.h"
#include "sphere.h"

void error(const char* message)
{
    perror(message);
    exit(1);
}

#include "colour.h"
#include "camera.h"
#include "material.h"

#include <iostream>

#include <string.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

const double ASPECT_RATIO = 1.0;
const unsigned int WIDTH = 256;
const unsigned int HEIGHT = static_cast<int>(WIDTH / ASPECT_RATIO);
const int SAMPLES_PER_PIXEL = 8;
const int MAX_DEPTH = 5;

// fee2aa
// 
const colour pink(254.0/255.0, 226.0/255.0, 170.0/255.0);
const colour grey(0.133, 0.133, 0.133);

colour ray_colour(const ray& r, const hittable& world, int depth)
{
    hit_record rec;
    if (depth <= 0)
    {
        return grey;
    }

    if (world.hit(r, 0.001, infinity, rec))
    {
        ray     scattered;
        colour  attenuation;

        if (rec.mat_ptr->scatter(r, rec, attenuation, scattered))
        {
            return attenuation * ray_colour(scattered, world, depth-1);
        }

        return grey;
    }

    vec3 unit_direction = unit_vector(r.direction());
    auto t = 0.5 * (unit_direction.y() + 1.0) + 0.5;

    return lerp(grey, pink, t);
}

hittable_list random_scene()
{
    hittable_list world;


    //auto ground_material = make_shared<lambertian>(pink);
    //world.add(make_shared<sphere>(point3(0,-1000,0), 1000, ground_material));

    //for (int a = -11; a < 11; a++)
    //{
    //    for (int b = -11; b < 11; b++)
    //    {
    //        auto choose_mat = random_double();
    //        point3 centre(a + 0.9*random_double(), 0.2, b + 0.9*random_double());

    //        if ((centre - point3(4, 0.2, 0)).length() > 0.9)
    //        {
    //            shared_ptr<material> sphere_material;

    //            if (choose_mat < 0.8)
    //            {
    //                // diffuse
    //                //auto albedo = colour::random() * colour::random();
    //                sphere_material = make_shared<lambertian>(pink);
    //                world.add(make_shared<sphere>(centre, 0.2, sphere_material));
    //            }
    //            else if (choose_mat < 0.95)
    //            {
    //                // metal
    //                auto fuzz = random_double(0, 0.5);
    //                sphere_material = make_shared<metal>(pink, fuzz);
    //                world.add(make_shared<sphere>(centre, 0.2, sphere_material));
    //            }
    //            else
    //            {
    //                // glass
    //                sphere_material = make_shared<dielectric>(1.5);
    //                world.add(make_shared<sphere>(centre,0.2, sphere_material));
    //            }
    //        }
    //    }
    //}

    auto material1 = make_shared<dielectric>(1.5);
    world.add(make_shared<sphere>(point3(0, 0, 0), 3.0, material1));

    //auto material2 = make_shared<lambertian>(pink);
    //world.add(make_shared<sphere>(point3(-4, 1, 0), 1.0, material2));
    
    auto material3 = make_shared<metal>(pink, 0.5);
    int sphere_count = 10;
    for (int i = 0; i < sphere_count; i++)
    {
        float a = 6.28 * (float)i/sphere_count - 100.0;
        float r = 8.0;
        float x = r*sin(a);
        float y = 2.0*cos(a);
        float z = r*cos(a);
        point3 pos(x,y,z);
        world.add(make_shared<sphere>(pos, 2.0, material3));
    }


    return world;
}

// file descriptor of the socket we're listening for connections on
//
// returns fd for the client connection
int accept_client(int sockfd)
{
    int newsockfd;
    struct sockaddr_in cli_addr;
    socklen_t clilen = sizeof(cli_addr);

    newsockfd = accept(sockfd, (struct sockaddr*)&cli_addr, &clilen);
    if (newsockfd < 0)
    {
        error("ERROR accepting client");
    }

    return newsockfd;
}

int wait_for_client(int& sockfd)
{
    int newsockfd;
    struct sockaddr_in serv_addr;

    // open socket and await connection from client
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0)
    {
        error("ERROR creating socket");
    }

    bzero((char*)&serv_addr, sizeof(serv_addr));

    // we successfully created the socket, configure it for binding
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = INADDR_ANY;
    serv_addr.sin_port = htons(64999);  // convert number from host to network byte order

    // this is a bit of developer QoL so we can iterate more quickly
    // TODO: make it possible to disable this debug/release build configuration
    const int enable = 1;
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int)) < 0)
    {
        error("ERROR setsockopt(SO_REUSEADDR) failed");
    }
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEPORT, &enable, sizeof(int)) < 0)
    {
        error("ERROR setsockopt(SO_REUSEPORT) failed");
    }

    // bind the socket
    if (bind(sockfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) < 0)
    {
        error("ERROR binding socket");
    }

    // successfully bound socket, start listening for connections
    listen(sockfd, 5);

    newsockfd = accept_client(sockfd);
    return newsockfd;
}

void send_message(int sock, const char* message)
{
    int written = write(sock, message, strlen(message));
    if (written < 0)
    {
        error("ERROR sending message to the client");
    }
    printf("SEND %s\n", message);
}

void send_image_dimensions(int sock, unsigned int width, unsigned int height)
{
    width = htonl(width);
    height = htonl(height);

    int written = write(sock, &width, sizeof(uint32_t));
    if (written < 0)
    {
        error("ERROR writing width");
    }
    written = write(sock, &height, sizeof(uint32_t));
    if (written < 0)
    {
        error("ERROR writing height");
    }
}

int main()
{
    int sockfd;
    int newsockfd = wait_for_client(sockfd);

    printf("got a connection!\n");

    send_image_dimensions(newsockfd, WIDTH, HEIGHT);

    //std::cout << "P3\n" << WIDTH << ' ' << HEIGHT << "\n255\n";

    hittable_list world = random_scene();

    auto dist_to_target = 10.0;
    auto dist_to_focus = dist_to_target + 1.0;
    auto cam_y = 1.0;
    point3 lookfrom(0,cam_y,-dist_to_target);
    point3 lookat(0,0,0);
    vec3 vup(0,1,0);
    auto aperture = 0.5;

    camera cam(lookfrom, lookat, vup, 47, ASPECT_RATIO, aperture, dist_to_focus);

    for (int j = HEIGHT - 1; j >= 0; --j)
    {
        std::cerr << "\rScanlines remaining: " << j << ' ' << std::flush;
        for (int i = 0; i < WIDTH; ++i)
        {
            colour pixel_colour(0,0,0);

            for (int s = 0; s < SAMPLES_PER_PIXEL; ++s)
            {
                auto u = (i + random_double()) / (WIDTH-1);
                auto v = (j + random_double()) / (HEIGHT-1);
                ray r = cam.get_ray(u, v);
                pixel_colour += ray_colour(r, world, MAX_DEPTH);
            }

            //write_colour_to_stream(std::cout, pixel_colour, SAMPLES_PER_PIXEL);
            write_colour_to_socket(newsockfd, pixel_colour, SAMPLES_PER_PIXEL);
        }
    }

    // close client socket
    close(newsockfd);

    printf("closed client connection\n");

    // close listening socket
    close(sockfd);

    printf("closed listening socket\n");

    printf("done!\n");

    return 0;
}

